(* Exercices de programmation du chapitre 1 *)

(* La syntaxe abstraite des expressions mini-ML *)

type expression =
    Var of string
  | Const of int
  | Op of string
  | Fun of string * expression
  | App of expression * expression
  | Paire of expression * expression
  | Fst of expression
  | Snd of expression
  | Let of string * expression * expression

(* Substitution d'une variable x par une expression b dans une expression a *)

let rec subst a x b =
  match a with
    Var v -> if v = x then b else a
  | Const n -> a
  | Op o -> a
  | Fun(v, a1) -> if v = x then a else Fun(v, subst a1 x b)
  | App(a1, a2) -> App(subst a1 x b, subst a2 x b)
  | Paire(a1, a2) -> Paire(subst a1 x b, subst a2 x b)
  | Fst a1 -> Fst(subst a1 x b)
  | Snd a1 -> Snd(subst a1 x b)
  | Let(v, a1, a2) -> Let(v, subst a1 x b, if v = x then a2 else subst a2 x b)

(* La fonction partielle d'valuation *)

exception RuntimeError

let rec valeur a =
  match a with
  | Var _ -> assert false (* l'expression doit tre close *)
  | Const c -> Const c
  | Op o -> Op o
  | Fun(x, a) -> Fun(x, a)
  | App(a1, a2) ->
      begin match valeur a1 with
        Fun(x, a) -> valeur (subst a x (valeur a2))
      | Op "+" ->
	  begin match valeur a2 with
	  | Paire(Const n1, Const n2) -> Const(n1 + n2)
          | _ -> raise RuntimeError (* arguments incorrects pour l'addition *)
          end
      |	_ -> raise RuntimeError (* arguments incorrects pour l'application *)
      end
  | Paire(a1, a2) -> Paire(valeur a1, valeur a2)
  | Fst a ->
      begin match valeur a with
      |	Paire(v1, _) -> v1
      |	_ -> raise RuntimeError (* argument incorrect pour la projection *)
      end
  | Snd a ->
      begin match valeur a with
      |	Paire(_,v2) -> v2
      |	_ -> raise RuntimeError (* argument incorrect pour la projection *)
      end
  | Let(x, a1, a2) -> valeur (subst a2 x (valeur a1))

(* L'algbre de types *)

type exprtype =
    Type_base of string
  | Type_var of string
  | Type_flche of exprtype * exprtype
  | Type_produit of exprtype * exprtype

type schma = { quantif: string list; corps: exprtype }

(* Calcul des variables libres *)

let rec union l1 l2 =
  match l1 with
    [] -> l2
  | hd :: tl -> union tl (if List.mem hd l2 then l2 else hd :: l2)

let rec variables_libres t =
  match t with
    Type_base b -> []
  | Type_var v -> [v]
  | Type_flche(t1, t2) -> union (variables_libres t1) (variables_libres t2)
  | Type_produit(t1, t2) -> union (variables_libres t1) (variables_libres t2)

let rec diffrence l1 l2 =
  match l1 with
    [] -> []
  | hd :: tl ->
      if List.mem hd l2 then diffrence tl l2 else hd :: diffrence tl l2

let variables_libres_schma s =
  diffrence (variables_libres s.corps) s.quantif

let rec variables_libres_env env =
  match env with
    [] -> []
  | (ident, schma) :: reste ->
      union (variables_libres_schma schma) (variables_libres_env reste)

(* Gnralisation d'un type en un schma *)

let gnralisation typ env =
  { quantif = diffrence (variables_libres typ) (variables_libres_env env);
    corps = typ }

(* Prdicat d'instanciation.
   typ est une instance de sch = forall a1...aN. tau
   s'il existe une substitution S sur les a1...aN telle que typ = S(tau).
   Ici, nous construisons la substitution incrmentalement 
   par rcurrence sur typ et tau *)

let instance typ sch =
  let subst = ref [] in
  let rec inst t1 t2 =
    match (t1, t2) with
      (Type_base b1, Type_base b2) -> b1 = b2
    | (Type_var v, _) when List.mem v sch.quantif ->
        begin try
          List.assoc v !subst = t2
        with Not_found ->
          subst := (v, t2) :: !subst; true
        end
    | (Type_var v, Type_var v') -> v = v'
    | (Type_flche(r1, s1), Type_flche(r2, s2)) ->
        inst r1 r2 && inst s1 s2
    | (Type_produit(r1, s1), Type_produit(r2, s2)) ->
        inst r1 r2 && inst s1 s2
    | (_, _) -> false
  in inst sch.corps typ

